// Time-stamp: <2006-11-07 12:07:18 uijw>
#ifndef MAPED_H__
#define MAPED_H__

#include <vector>
#include <list>
#include <algorithm>
#include "snode.H"
#include "gsl_rand.H"
#include "availableMap.H"

class availInfo;

class mpforest {
  friend class snode;
  typedef std::vector<snode *>::iterator itor;
  typedef std::vector<snode *>::const_iterator citor;

public:
  // random constructor for shape (without genotypes)
  mpforest(int sampleSize,double self, int N,int T,rng &r);
  // constructor for a set of genotypes using a sequential sampler
  mpforest(const std::vector<genotype> &gg,double self, int N,int T, const popFreq &pf, rng &r);
  
  // constructor from an input stream      
  mpforest(std::istream &in);
  /// destructor
  ~mpforest() {
    std::for_each(sample.begin(),sample.end(),DeleteObject());
    std::for_each(ancestors.begin(),ancestors.end(),DeleteObject());

  }

  void gibbs_sample(int T, double N, double self, const popFreq &pf, rng &r);
  bool metro_addremove_k(int kk,int T, double N, double self, const popFreq &pf, rng &r);
  
  void check_add(double self, double N, int T
		 , const popFreq &pf, rng &r);

  /** Returns the numbr of samples                                          */
  int nsamples() const {
    return sample.size();
  }
  /** returns the number of ancestors                                       */
  int nancestors() const {
    return ancestors.size();
  }
  /** returns the number of ancestors                                       */
  int nroots() const {
    return available.back().size();//roots.size();
  }
  
  int whichavailable(snode * a, int gen) {
    itor i=std::find(available[gen].begin(),available[gen].end(),a);
    if (i==available[gen].end())
      throw std::domain_error("This is not available in whichavailable");
    return std::distance(available[gen].begin(),i);
  }

  itor rootbegin() {
    return available.back().begin();
  }
  itor rootend() {
    return available.back().end();
  }
  citor rootbegin() const{
    return available.back().begin();
  }
  citor rootend() const {
    return available.back().end();
  }


  // how many available lines at time Time+1
  size_t nleft(size_t Time) const {
    assert(Time<available.size());
    return available[Time].size();
  }
  /** Return the value of T used in the analysis                   */
  int getT() const {
    return available.size();
  }
  /** Remove a list of ancestors from the ancestor vector                   */
  void removeanclist(std::list<snode *> &li) {
    std::list<snode *>::const_iterator ii=li.begin();
    while(ii!=li.end()) {
      ancestors.erase(std::remove(ancestors.begin(),ancestors.end(),*ii)
		      ,ancestors.end()
		      );
      ii++;
    }
  }
  /** add a sample without any reference to the genotype   */
  snode *add_sample(int lab, double s, int N, int T, rng &r);
  /** mutate the sample                                    */
  popFreq mutate(int loci, double theta, rng &r);
  void mutate(const popFreq &pf,rng &r);

  std::ostream &printGenotypes(std::ostream &o);
  double peel(double self,const popFreq &pf,bool repeel=true);
  int countOutcrosses();


  /** What is the probability of the selfing (and non selfing) for the forest     */
  double lpselfing(double self) const;

  /** Add a sample that has a genotype for a given selfing rate, T etc.        */
  double add_geno_sample(snode *samp,double s, double N, int T
			 , const popFreq &pf, rng &r);
  
  /** remove the last sample - returns a pointer to the root that is altered (or 0 if removed)  
  */
  snode *remove_last_sample(availInfo &AI);

  snode *operator[](int i) {
    return sample[i];
  }

  void swaplast(int i) {
    snode *tmp=sample.back();
    sample.back()=sample[i];
    sample[i]=tmp;
  }
 
   // print the trees
  std::ostream &printTrees(std::ostream &o, bool printGenotypes=false
			   ,bool singletons=false,bool printSelfing=true);

  /** get the genotype data (after a simulation                                          */
  std::vector<genotype> extractData() const;
  void matchLast(int Tim, const std::vector<snode *> &lines);

  void fixAvailable(const std::vector<snode *> &tofix);
  void fixAvailable(snode *tofix);

  double lprobtrees(double s, double N, int T);

  snode *roots(int i) {
    return available.back()[i];
  }
  bool check_trees(const popFreq &pf) const;
  /** What is the minimum value that N can be ?                        */
  int minN() const {
    return nleft(0);
  }
  std::vector<int> partition(int gen=-1);
  std::vector<int> partitionMROA();
  
  void reorder(rng &r) {
    permute(sample,r);
  }
  void removeNodesReverse(const std::vector<snode *> &NR, const popFreq &pf);

private:
  // std::vector<snode *> roots;
  std::vector<snode *> ancestors;
  std::vector<snode *> sample;
  std::vector<int> genBefore;

  //available refers to the available snodes for coalescence in the generation above time
  std::vector<std::vector<snode *> > available;


  // Private functions
  std::vector<double> calculateP(const genotype &g, double self, double N, int T
					  ,const popFreq &pf);
  snode *insertsample(snode *samp,const  availInfo &a);
  void testAvailable(snode *tofix);
};

#endif
