#ifndef OPTWRAPB_H__
#define OPTWRAPB_H__

#include <string>
#include <map>
#include <iostream>
#include <vector>
#include <sstream>
#include <stdexcept>


// write a command line parser than returns a list of
// options and text associated with the options

// class clparser {
//   clparser(int argc, char *argv[]);

// private:
//   std::map<const char *,std::vector<const char *> >
// };


class options_exception: public std::out_of_range {
public:
  options_exception(const std::string &err):
    out_of_range(err) {
  };
};

void Tokenise(const std::string& str,
                      std::vector<std::string>& tokens,
	      const std::string& delimiters = " ");


class optBase {
public:
  optBase(const char *desc)
    :description(desc) {};
  virtual ~optBase() {};
  virtual std::ostream &print(std::ostream &o,const char *before="",const char *after="")=0;
  virtual std::ostream &printdesc(std::ostream &o) {
    return o << description;
   }
  virtual void switchval()=0;
  virtual void read(const std::string &val)=0;
protected:
  std::string description;
};


template <typename T>
class opt: public optBase {
public:
  opt(T *val,const char *desc, const T &def)
    :optBase(desc),x(val) {
    *x=def;
  }
  virtual void switchval() {
    throw options_exception("trying to switch an option that expects a value");
  }
  virtual std::ostream &print(std::ostream &o,const char *before="",const char *after="") {
    o << before << *x << after;
    return o;
  };
  virtual void read(const std::string &val)
  {
    std::istringstream iss(val);
    iss >> *x;
  }
      
  opt(T *val);
private:
  T *x;      // The actual value, somewhere else in the program
};


class switchopt: public optBase {
public:
  switchopt(bool *val,const char *desc)
    :optBase(desc),x(val) {
    *x=false;
  }
  virtual void switchval() { 
    *x=true;
  }
  std::ostream &print(std::ostream &o,const char *before="",const char *after="") {
    o << before;
    if (*x) o << "true";
    else o << "false";
    o << after;
    return o;
  }
  virtual void read(const std::string &val) {
   throw options_exception("trying to read an option that is just a switch\n");
 }
private:
  bool *x;      // The actual value, somewhere else in the program
};


class options {
  typedef std::map<std::string, optBase *>::iterator itor;
  typedef std::map<std::string, optBase *>::const_iterator citor;
  typedef std::vector<optBase *>::iterator pos_itor;
public:
  options(const char *text, const char *vers="0.01")
    :description(text),version(vers) {
  }
  template <typename T> 
  void add(T *val,const char *text,const char *desc, const T &def) {
    optmap[text]=new opt<T>(val,desc,def);
  } 
  // specialise for the string
  void add(std::string *val,const char *text,const char *desc, const char *def) {
    optmap[text]=new opt<std::string>(val,desc,std::string(def));
  }
  // specialise for a bool
  void add(bool *val,const char *text,const char *desc) {
    optmap[text]=new switchopt(val,desc);
  }
  // add a positional option
  template <typename T> 
  void add(T *val, const char *text, const T &def) {
    std::ostringstream oss;
    oss << "Position " << posOpt.size()+1 << " " << text;
    posOpt.push_back( new opt<T>(val,oss.str().c_str(),def) );
  }
  // specialised for string
  void add(std::string *val, const char *text,const char *def) {
    std::ostringstream oss;
    oss << "Positional Option " << posOpt.size()+1 << ": " << text;
    posOpt.push_back( new opt<std::string>(val,oss.str().c_str(),def) );
  }

  void readcommandline(int argc, char *argv[]);
  void readfromfile(const char *filename);
   
  std::ostream &print(std::ostream &o) const {
    printversion(o,"","#");
    citor i=optmap.begin();
    while (i!=optmap.end()) {
      o << i->first << " = ";
      i->second->print(o);
      o  << std::endl;
      i++;
    }
    return o;
  }
  void printversion(std::ostream &o, const char* progname,const char *start="") const {
    o << start << progname  << " version " << version << "  (compiled on " 
      << __DATE__<<" at " << __TIME__ <<")" << std::endl;
  }
  void printhelp(std::ostream &o);
private:
  std::map<std::string, optBase *> optmap;
  std::vector<optBase *> posOpt;   // positional options
  std::string description,version;
};


std::ostream &operator<<(std::ostream &o, const options &opt);

#endif
