// Time-stamp: <2006-10-17 17:15:23 uijw>
/** structure for keeping track of probabilities as we peel a tree where
 * we have some outcrossing, but no loops                              
 * The map keeps track of the legal set genotypes that could have led to 
 * the descendents seen - including completely missing ones that may be 
 * possible for outcrossing events  
 */
#ifndef PROBMAP_H__
#define PROBMAP_H__

#include "genotype.H"
#include "popFreq.H"

  
/** This class holds an upper matrix                                 */
class Peel {
  friend class PeelVector;
public:
  /** Constructor for the Peeling class                               */
  Peel(int nalleles):alleles_(nalleles), p_(alleles_*alleles_,0.0) {
  };
  Peel() {};
  /** assignment operator                                             */
  Peel & operator=(const Peel &rhs) {
    if (this != &rhs) {
      p_=rhs.p_;
      alleles_=rhs.alleles_;
    }
    return *this;
  }
  /** Calculate the U and V coeficients based on Kevin's equations   */
  std::pair<double,double> UV(const std::vector<double> &popf) const;

  /** Allocate the necessary space for the vector                     */
  void allocate(int nalleles,double val=0.0) {
    alleles_=nalleles;
    p_.assign(alleles_*alleles_,val);
  }
  /** returns the value at [a1][a2] for unordered a1 and a2           */
  double at(int a1, int a2) const {
    if (a1<a2) return ordered_at(a1,a2);
    else return ordered_at(a2,a1);
  }
  /** Change the value at [a1][a2] for unordered a1 and a2            */
  double &at(int a1, int a2) {
    if (a1<a2) return ordered_at(a1,a2);
    else return ordered_at(a2,a1);
  }

  double &ordered_at(int a1, int a2)
  {
    assert(a2>=a1);
    return p_[a1*alleles()+a2];
  }
  double ordered_at(int a1, int a2) const 
  {
    assert(a2>=a1);
    return p_[a1*alleles()+a2];
  }
  /* The number of alleles for this peel                                      */
  int alleles() const {
    return alleles_;
  }
  /** Add a selfing leaf                                */
  void SelfingLeaf(const char g[2]);
  /** what are the probabilities for a selfing node with single descendent 
      with probabilities given by a                                        */ 
  void SelfingNode(const Peel &a);
  /** Add an outcrossed leaf with genotype g with popfreqs pf             */
  void OutcrossLeaf(const char g[2],const std::vector<double> &freq);
  /** what are the probabilities for an outcrossed node with single descendent 
      with probabilities given by a                                        */ 
  void OutcrossNode(const Peel &a,const std::vector<double> &freq );

  /** is this Peel compatible with p2                                        */
  bool compatible(const Peel &p2) const;
 
  /** print out the upper matrix of probabilities         */
  std::ostream &print(std::ostream &o) const;

  void copy(const Peel &p2) {
    assert(p_.size()==p2.p_.size());
    std::vector<double>::iterator i=p_.begin();
    std::vector<double>::const_iterator j=p2.p_.begin();
    while (i!=p_.end()) {
      *(i++) = *(j++);
    }
  }
  /** Is this upper matrix idnetically zero ?                                 */
  bool is_zero(double mn=1E-12) {
    for (int i=0;i<alleles();i++)
      for (int j=i;j<alleles();j++) 
	if (fabs(ordered_at(i,j)) > mn)
	  return false;
    return true;
  }

  /** muliply Peel by p2 - element by element                                 */
  void multiply(const Peel &p2) {
    assert(p_.size()==p2.p_.size());
    std::vector<double>::iterator i=p_.begin();
    std::vector<double>::const_iterator j=p2.p_.begin();
    while (i!=p_.end()) {
      *(i++) *= *(j++);
    }
  }
  /** set to a genotype (for non tree models)       */
  void setgeno(const char g[2]) {
    at(g[0],g[1])=1.0;
  }

  bool equals(const Peel &p2) const;
private:
  int alleles_;
  std::vector<double> p_;


};

/** A class to hold a vector of matrices of probability maps  
 *
 *
 *
 *                                                               */
class PeelVector {
public:
  /** Constructor                                             */
  PeelVector(const std::vector<int> &alleles) {
    allocate(alleles,0.0);
  }
  /** constuctor for data (no outcrossing or selfing         */
  PeelVector(const genotype &g,const std::vector<int> &alleles) {
    allocate(alleles,0.0);
    for (size_t i=0;i<alleles.size();i++) 
      pv[i].setgeno(g[i]);
  }

  /** The default constructor                                 */
  PeelVector() {};

  

  void OutcrossLeaf(const genotype &g,const popFreq &pf) {
   for (size_t locus =0;locus <pv.size();locus++) {
      pv[locus].OutcrossLeaf(g[locus],pf[locus]);
    }
  }
 void OutcrossNode(const PeelVector &a,const popFreq &pf) {
   for (size_t locus =0;locus <pv.size();locus++) {
      pv[locus].OutcrossNode(a.pv[locus],pf[locus]);
    }
  }
  void SelfingLeaf(const genotype &g) {
    for (size_t locus =0;locus <pv.size();locus++) {
      pv[locus].SelfingLeaf(g[locus]);
    }
  }
  void SelfingNode(const PeelVector &a) {
    for (size_t locus =0;locus <pv.size();locus++) {
      pv[locus].SelfingNode(a.pv[locus]);
    }
  }

  /** Allocate space for the vector                          */
  void allocate(const std::vector<int> &alleles,double val=0.0) {
    pv.resize(alleles.size());
    for (size_t i=0;i<pv.size();i++) 
      pv[i].allocate(alleles[i],val);
  }

  /** multiple this by PeelVector b (by locus)               */
  void multiply(const PeelVector &b) {
    for (size_t i=0;i<pv.size();i++) 
      pv[i].multiply(b.pv[i]);
  }

  /** multiple this by PeelVector b (by locus)               */
  bool compatible(const PeelVector &b) {
    for (size_t i=0;i<pv.size();i++) 
      if (!pv[i].compatible(b.pv[i])) return false;
    return true;
  }

  /** How many loci?                                         */
  size_t nloci() const {
    return pv.size();
  }

  /** is the upper matrix identically  zero at any locus?                */
  bool is_zero(double mn=1E-16) {
    for (size_t i=0;i<pv.size();i++) 
      if (pv[i].is_zero(mn)) return true;
    return false;

  }
  
  /** print out the vector of peeling probabilities           */
  std::ostream &print(std::ostream &o) const;
  /** print out the vector of peeling probabilities           */
  std::ostream &print(std::ostream &o, size_t locus) const {
    pv[locus].print(o);
    return o;
  }

  /** calculate the U,V pair for locus                        */
  std::pair<double,double> UV(const std::vector<double> &popf,size_t locus) const {
    assert(locus<nloci());
    return pv[locus].UV(popf);
  }

  void copy(const PeelVector &b) {
    assert(pv.size()==b.nloci());
    for (size_t i=0;i<pv.size();i++) 
      pv[i].copy(b.pv[i]);
  }
  bool equals(const PeelVector &b) const {
    for (size_t i=0;i<pv.size();i++) 
      if (!pv[i].equals(b.pv[i])) return false;
    return true;
  }

private:
  std::vector<Peel> pv;
};

std::ostream &operator<<(std::ostream &o, const PeelVector &pv);

#endif
