// @file
// Time-stamp: <2006-11-07 13:12:04 uijw>

#ifndef SNODE_H__
#define SNODE_H__
#include <vector>
#include <iostream>
#include "genotype.H"
#include "popFreq.H"
#include "probMap.H"
      
/** Selfing nodes 
 * A class to hold information on Selfing trees
 * and to keep partial information     */

class snode { 
  friend class mpforest;
  typedef std::vector<snode *>::iterator itor; 
  typedef std::vector<snode *>::const_iterator citor; 

public:
  /** Constructors                                        */
  snode(int l)
    :ancestor_(0),time_(0),label_(l),outcross(false) {
  }
  snode(int t, snode *d,bool oc=false)
    :ancestor_(0),time_(t),label_(-1),outcross(oc) {
    descendent_.push_back(d);
    d->ancestor_=this;
  }
  snode(const genotype &gg,int l) 
    :ancestor_(0),time_(0),label_(l),outcross(false),g(gg) {
  }
  /** destructor       */
  ~snode() {  };
 /** Add a descendent node to the tree                  */
  void add(snode *aa) {
    descendent_.push_back(aa);
    aa->ancestor_=this;
  }
  /** How many descendents does the node have? */
  size_t ndesc() const {
    return descendent_.size();
  }  
  /** What are the total number of descendents?   */
  int total_descendents() {
    if (ndesc()==0) return 1; // leaf
    int count=0;
    itor ii=descendent_.begin();
    while (ii!=descendent_.end()) {
      count+= (*ii++)->total_descendents();
    }
    return count;
  } 
  /** return descendent i         */
  snode *descendent(size_t i) {
    assert(i>=0&&i<descendent_.size());
    return descendent_[i];
  }
  /** Erase one of the descendents of the node   */
  void erasedescendent(snode *d) {
    assert(descendent_.size()>0);
    itor i = std::remove(descendent_.begin(),descendent_.end(),d);
    assert(i != descendent_.end());
    // so the node to be removed is *i
    descendent_.erase(i,descendent_.end());
  }
  /** Which node is the ancestor of the current node? */
  snode *up() const {
    return ancestor_;
  }
  void up(snode *val){  
    ancestor_=val;
  }
  /** The most recent outcrossed ancestor of this node 
   * (could be this node if it is itself outcrossed)      */
  snode *MROA()  {
    snode *a=this;
    for (;;) {
      if (a->outcross||a->up()==0) return a;
      a=a->up();
    }
  }
  /** What is hte genotype of this node ?             */
  const genotype &geno() const {
    return g;
  };
  /** What is the time of this node                   */
  double time() const {
    return time_;
  }
  /** print a line about the status                   */
   std::string status() const {
     if (outcross) return "outcross";
     else return "selfing";
   }
     /** print a character about the status           */
   char shortstatus() const {
     if (outcross) return 'O';
     else return 'S';
   }
  /** Find the time of the next node that is not 
   * a selfing node                                   */
  int nexttime() {
    int count=1;
    snode *a=this;
    while (a->ndesc()==1) {
      a=a->descendent_[0];
      count++;
    }
    return count;
  }
  /** find the root                                   */
  snode *find_root() {
    snode *a=this;
    while(a->up()!=0) a=a->up();
    return a;
  }
  /** go up n generations from this node              */
  snode *up_n(int n) {
    snode *a=this;
    for (int i=0;i<n;i++) {
      if (a->up()==0) 
	throw std::domain_error("Cannot go up here in up_n");
      a=a->up();
    }
    return a;
  }
  /** Random drift at a node                          */
  void drift(rng &r) {
    for (size_t  i=0;i<g.nloci();i++) {
      double p=r();
      if (p<0.25) g(i,0)=g(i,1);
      else if (p<0.5) g(i,1)=g(i,0);
    }
  } 
  /* Return the labal at this node                   */
  int label() const {
    return label_;
  }

  void recurse_getgenotypes(const popFreq &pf,rng &r);

  void peel(const popFreq &pf);
  void recursive_peel(const popFreq &pf);
  
  double pjoin(double self,const popFreq &pf, const PeelVector &pv);
  void fixProbVector();

  std::ostream &print_newick(std::ostream &o,bool printGenotype=false,bool printSelfing=true);
  std::string selfstring();

  bool check(const popFreq &pf ) const;

private:
  snode *ancestor_;
  std::vector<snode *> descendent_;
  int time_;
  int label_;
  bool outcross;
  genotype g;
  PeelVector p;
};

std::ostream &operator<<(std::ostream &o, const genotype &g);
#endif


